package mms.mm1;

import java.io.File;
import java.io.IOException;

import com.objectxp.mms.MMSDeliveryReport;
import com.objectxp.mms.MMSException;
import com.objectxp.mms.MMSListener;
import com.objectxp.mms.MMSNotification;
import com.objectxp.mms.MMSReadReport;
import com.objectxp.mms.MMSStatus;
import com.objectxp.mms.MMSService;
import com.objectxp.mms.MMSServiceFactory;
import com.objectxp.mms.message.MMSMessage;
import com.objectxp.mms.protocol.MM1Service;
import com.objectxp.mms.protocol.ProtocolException;
import com.objectxp.mms.transport.TransportException;
import com.objectxp.msg.GsmSmsService;
import com.objectxp.msg.MessageEvent;
import com.objectxp.msg.MessageEventListener;
import com.objectxp.msg.MessageException;
import com.objectxp.msg.MultiPartReceiver;
import com.objectxp.msg.SmsService;


/**
 * This example demonstrates how to receive a MultiMediaMessage (MMS) over MM1.
 * 
 * The MMS Relay/Server sends a notification using WAP-Push (SMS) to indicate 
 * incoming messages. Therefore, a GsmSmsService will be used to receive those
 * notifications.
 *  
 */
public class MM1ReceiveExample implements MMSListener, MessageEventListener
{
  private SmsService service;
  private MMSService mmsService;
  public static void main(String[] args) throws Exception
  {
    
    if(args.length < 1 ) {
      System.err.println("Usage: MM1ReceiveExample <configfile> [service-name]");
      System.exit(1);
    }
    
    MM1ReceiveExample receiver = new MM1ReceiveExample(args[0], args.length>1?args[1]:null);
    
    // Start receiving SMS
    receiver.startGsmService();
    
    // Wait for key-press
    while( System.in.available() < 1 ) {
      Thread.sleep(1000);
    }
    
    receiver.shutdown();
  }

  public MM1ReceiveExample(String configFile, String serviceName) throws IOException, MessageException 
  {
    // Construct a MMS Service and set the Listener for this service:
    MMSServiceFactory factory = MMSServiceFactory.createFactory(new File(configFile));
    mmsService = factory.getService(serviceName);
    mmsService.setListener(this);
    
    // Create and initialize the SMS Service
    service = new GsmSmsService();
    System.out.print("Initializing GSM device...");
    service.init(new File(configFile)); 
    System.out.println("OK");
    
    // Since MMS notifications come in multiple fragments, we engage the
    // MultiPartReceiver for assembling them.
    MultiPartReceiver mpr = new MultiPartReceiver(360*1000, 50, this);
    service.addMessageEventListener(mpr);
    
    // The MM1 service can be used to intercept MMS messages delivered over
    // WAP-PUSH (SMS). 
    if(mmsService instanceof MM1Service) {
      System.out.println("Registering MM1 service as SMS event listener.");
      mpr.addMessageEventListener((MessageEventListener) mmsService);
    }
    
  }

  
  private void startGsmService() throws IOException, MessageException
  {
    System.out.print("Connecting to GSM device...");
    service.connect();
    System.out.println("OK");

    System.out.print("Start receiving Messages...");
    service.startReceiving();
    System.out.println("OK");
  }

  private void shutdown() throws MMSException, ProtocolException, TransportException, IOException {
    if( service != null ) {
      service.destroy();
    }
    
    if( mmsService != null ) {
      mmsService.disconnect();
    }
  }

  public void handleMessageEvent(MessageEvent event) 
  {
    System.out.println("--- SMS Message event received ---");
    System.out.println(event);
  }
  
  public void handleNotification(MMSService src, MMSNotification notification) 
  {
    System.out.println("--- MMS Notification received ---");
    System.out.println(notification);
    System.out.println("---------------------------------");
      
      // Disconnect the GSMService
      if (service != null){
        try {
          System.out.print("Stopping receiption of SMSes...");
          service.stopReceiving();
          System.out.println("OK");
          System.out.print("Disconnecting from GSM device...");
          service.disconnect();
          System.out.println("OK");
        } catch (MessageException e) {
          System.err.println("disconnection SmsService failed.");
          e.printStackTrace();
        } 
      }      

      try {
        // Connect to the MMSC
        System.out.print("Connecting to MMSC...");
        mmsService.connect();
        System.out.println("OK");
        
        // Fetch the MMS
        System.out.print("Fetching MMS...");
        mmsService.fetch(notification);
        System.out.println("OK");
        
      } catch (Exception e) {
        System.err.println("could not fetch MMSMessage with id "+notification.getTransactionId());
        e.printStackTrace();
      } finally {
        // Disconnect from MMSC
        if (mmsService!=null) {
          try  {
            System.out.print("Disconnecting from MMSC...");
            mmsService.disconnect();
            System.out.println("OK");
          } catch (Exception e) {
            System.err.println("Could not disconnect mmsService after fetching."+e.getMessage());
            e.printStackTrace();
          }
        }
      }
      
      // reconnect GSM service for further messages
      try {
        startGsmService();
      } catch (Exception e) {
        System.err.println("could not reconnect GsmSmsSerice.");
        e.printStackTrace();
      }
  }
  
  public MMSStatus handleIncommingMessage(MMSService service, MMSMessage message) {
    System.out.println("--- MMS Message received ---");
    System.out.println(message);
    System.out.println("----------------------------");
    return MMSStatus.SUCCESS;
  }

  public MMSStatus handleDeliveryReport(MMSService service, MMSDeliveryReport report) {
    System.out.println("--- MMS Delivery Report received ---");
    System.out.println(report);
    System.out.println("------------------------------------");
    return MMSStatus.SUCCESS;
  }

  public MMSStatus handleReadReport(MMSService service, MMSReadReport report) {
    System.out.println("--- MMS Read Report received ---");
    System.out.println(report);
    System.out.println("--------------------------------");
    return MMSStatus.SUCCESS;
  }
}
